/**
 * Register domain for Apple Pay
 *
 * @module package/quiqqer/payment-stripe/bin/controls/backend/settings/ApplePayRegister
 * @author www.pcsg.de (Patrick Müller)
 */
define('package/quiqqer/payment-stripe/bin/controls/backend/settings/ApplePayRegister', [

    'qui/controls/Control',
    'qui/controls/loader/Loader',
    'qui/controls/buttons/Button',

    'Locale',
    'Ajax',

    'css!package/quiqqer/payment-stripe/bin/controls/backend/settings/ApplePayRegister.css'

], function (QUIControl, QUILoader, QUIButton, QUILocale, QUIAjax) {
    "use strict";

    var lg = 'quiqqer/payment-stripe';

    return new Class({

        Extends: QUIControl,
        Type   : 'package/quiqqer/payment-stripe/bin/controls/backend/settings/ApplePayRegister',

        Binds: [
            '$refresh',
            '$onImport',
            '$onRegisterBtnClick',
            '$getApplePayData',
            '$registerApplePayDomain'
        ],

        initialize: function (options) {
            this.parent(options);

            this.$Content = null;
            this.Loader   = new QUILoader();

            this.addEvents({
                onCreate: this.$onCreate,
                onResize: this.$onResize,
                onImport: this.$onImport
            });
        },

        /**
         * Control event: onImport
         */
        $onImport: function () {
            var Input  = this.getElm();
            Input.type = 'hidden';

            var Wrapper = new Element('div', {
                'class': 'quiqqer-payment-stripe-applepay field-container-field',
                html   : '<div class="quiqqer-payment-stripe-applepay-content"></div>'
            }).inject(Input, 'after');

            new Element('h2', {
                'class': 'quiqqer-payment-stripe-applepay-header',
                html   : QUILocale.get(lg, 'controls.backend.settings.ApplePayRegister.header')
            }).inject(Wrapper, 'top');

            this.$Content = Wrapper.getElement('.quiqqer-payment-stripe-applepay-content');
            this.Loader.inject(Wrapper);

            this.$refresh();
        },

        /**
         * Refresh register status display
         */
        $refresh: function () {
            var self = this;

            this.Loader.show();

            var registerIcon = 'fa fa-check',
                cancelIcon   = 'fa fa-close',
                registerText = QUILocale.get(lg, 'controls.backend.settings.ApplePayRegister.btn.register'),
                cancelText   = QUILocale.get(lg, 'controls.backend.settings.ApplePayRegister.btn.cancel');

            this.$getApplePayData().then(function (Domains) {
                self.$Content.set('html', '');

                for (var domain in Domains) {
                    if (!Domains.hasOwnProperty(domain)) {
                        continue;
                    }

                    var ListElm = new Element('li', {
                        'class': 'quiqqer-payment-stripe-applepay-entry',
                        html   : '<span>' + domain + '</span>' +
                            '<div class="quiqqer-payment-stripe-applepay-entry-btn"></div>'
                    }).inject(self.$Content);

                    var isRegistered = Domains[domain];

                    if (isRegistered) {
                        ListElm.addClass('quiqqer-payment-stripe-applepay-entry-registered');
                    } else {
                        ListElm.addClass('quiqqer-payment-stripe-applepay-entry-not-registered');
                    }

                    new QUIButton({
                        text        : isRegistered ? cancelText : registerText,
                        textimage   : isRegistered ? cancelIcon : registerIcon,
                        isRegistered: isRegistered,
                        domain      : domain,
                        events      : {
                            onClick: self.$onRegisterBtnClick
                        }
                    }).inject(ListElm.getElement('.quiqqer-payment-stripe-applepay-entry-btn'));
                }

                self.Loader.hide();
            });
        },

        /**
         * (Un)register Apple Pay domain
         *
         * @param {Object} Btn
         */
        $onRegisterBtnClick: function (Btn) {
            var self = this;

            this.Loader.show();

            this.$registerApplePayDomain(
                Btn.getAttribute('domain'),
                Btn.getAttribute('isRegistered')
            ).then(function (success) {
                self.Loader.hide();

                if (success) {
                    self.$refresh();
                }
            });
        },

        /**
         * Register a domain with Apple Pay
         *
         * @param {String} domain
         * @param {Boolean} [unregister]
         * @return {Promise}
         */
        $registerApplePayDomain: function (domain, unregister) {
            return new Promise(function (resolve, reject) {
                QUIAjax.post('package_quiqqer_payment-stripe_ajax_backend_settings_registerApplePayDomain', resolve, {
                    'package' : lg,
                    domain    : domain,
                    unregister: unregister ? 1 : 0,
                    onError   : reject
                });
            });
        },

        /**
         * Fetch Apple Day registration data for current system
         *
         * @return {Promise}
         */
        $getApplePayData: function () {
            return new Promise(function (resolve, reject) {
                QUIAjax.get('package_quiqqer_payment-stripe_ajax_backend_settings_getApplePayData', resolve, {
                    'package': lg,
                    onError  : reject
                });
            });
        }
    });
});
