<?php

namespace QUI\ERP\Payments\Stripe;

use QUI;
use Stripe\Event as StripeEvent;
use Stripe\Refund as StripeRefund;
use Stripe\Subscription as StripeSubscription;
use QUI\ERP\Accounting\Payments\Transactions\Handler as TransactionsHandler;
use QUI\ERP\Payments\Stripe\PaymentMethods\Recurring\Subscriptions;

/**
 * Class WebhookHandler
 *
 * Handles all relevant Stripe webhook events
 */
class WebhookHandler
{
    /**
     * General entry point for all Stripe webhook events
     *
     * @param StripeEvent $Event
     * @return void
     */
    public static function handleEvent(StripeEvent $Event)
    {
        switch ($Event->type) {
            case StripeEvent::CHARGE_REFUNDED:
                self::onChargeRefunded($Event);
                break;

            case StripeEvent::CHARGE_REFUND_UPDATED:
                self::onChargeRefundUpdated($Event);
                break;

            case StripeEvent::CUSTOMER_SUBSCRIPTION_DELETED:
                self::onCustomerSubscriptionDeleted($Event);
                break;

            case StripeEvent::CUSTOMER_SUBSCRIPTION_UPDATED:
                self::onCustomerSubscriptionUpdated($Event);
                break;
        }
    }

    /**
     * Stripe webhook: customer.subscription.deleted
     *
     * @param StripeEvent $Event
     * @return void
     */
    protected static function onCustomerSubscriptionDeleted(StripeEvent $Event)
    {
        Subscriptions::setSubscriptionAsInactive($Event->data->object->id);
    }

    /**
     * Stripe webhook: customer.subscription.updated
     *
     * @param StripeEvent $Event
     * @return void
     */
    protected static function onCustomerSubscriptionUpdated(StripeEvent $Event)
    {
        if ($Event->data->object->status !== StripeSubscription::STATUS_ACTIVE) {
            Subscriptions::setSubscriptionAsInactive($Event->data->object->id);
        }
    }

    /**
     * Stripe webhook: charge.refund.updated
     *
     * @param StripeEvent $Event
     * @return void
     */
    protected static function onChargeRefundUpdated(StripeEvent $Event)
    {
        if (empty($Event->data->object->metadata->refundTxId)) {
            return;
        }

        try {
            $RefundTransaction = TransactionsHandler::getInstance()->get(
                $Event->data->object->metadata->refundTxId
            );

            switch ($Event->data->object->status) {
                case StripeRefund::STATUS_SUCCEEDED:
                    $RefundTransaction->complete();
                    break;

                case StripeRefund::STATUS_CANCELED:
                case StripeRefund::STATUS_FAILED:
                    $RefundTransaction->error();
                    break;
            }
        } catch (\Exception $Exception) {
            QUI\System\Log::writeException($Exception);
        }
    }

    /**
     * Stripe webhook: charge.refunded
     *
     * @param StripeEvent $Event
     * @return void
     */
    protected static function onChargeRefunded(StripeEvent $Event)
    {
        if (empty($Event->data->object->refunds[0]->metadata->refundTxId)) {
            return;
        }

        try {
            $RefundTransaction = TransactionsHandler::getInstance()->get(
                $Event->data->object->refunds[0]->metadata->refundTxId
            );

            if ($RefundTransaction->isPending()) {
                $RefundTransaction->complete();
            }
        } catch (\Exception $Exception) {
            QUI\System\Log::writeException($Exception);
        }
    }
}
